-- Broken Bottle - 7DRL
-- Copyright (C) 2011 Darren Grey
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "engine.class"
require "engine.Actor"
require "engine.Autolevel"
require "engine.interface.ActorTemporaryEffects"
require "engine.interface.ActorLife"
require "engine.interface.ActorProject"
require "engine.interface.ActorLevel"
require "engine.interface.ActorStats"
require "engine.interface.ActorTalents"
require "engine.interface.ActorResource"
require "engine.interface.ActorFOV"
require "mod.class.interface.Combat"
require "engine.interface.ActorInventory"
local Map = require "engine.Map"

module(..., package.seeall, class.inherit(
	engine.Actor,
    engine.interface.ActorInventory,
	engine.interface.ActorTemporaryEffects,
	engine.interface.ActorLife,
	engine.interface.ActorProject,
	engine.interface.ActorLevel,
	engine.interface.ActorStats,
	engine.interface.ActorTalents,
	engine.interface.ActorResource,
	engine.interface.ActorFOV,
	mod.class.interface.Combat
))

function _M:init(t, no_default)
	-- Define some basic combat stats
	self.combat_armor = 0
	self.crit_defence = 0
	t.max_actions = t.max_actions or 100
	-- Default regen
	t.stamina_regen = t.stamina_regen or 1
	t.life_regen = 0 -- No life regen
	t.actions_regen = t.actions_regen or 100 -- Action Points reset to full each turn.  Don't give actors bonus regen
	-- Default melee barehanded damage
	self.combat = { dam=0, crit_chance = 0 }
   
	engine.Actor.init(self, t, no_default)
	engine.interface.ActorTemporaryEffects.init(self, t)
	engine.interface.ActorLife.init(self, t)
	engine.interface.ActorProject.init(self, t)
	engine.interface.ActorTalents.init(self, t)
	engine.interface.ActorResource.init(self, t)
	engine.interface.ActorStats.init(self, t)
	engine.interface.ActorLevel.init(self, t)
	engine.interface.ActorFOV.init(self, t)
   engine.interface.ActorInventory.init(self, t)
   
end

function _M:act()
	if not engine.Actor.act(self) then return end

	self.changed = true
	self:cooldownTalents()
	-- Regen resources
	self:regenLife()
	self:regenResources()
	-- Compute timed effects
	self:timedEffects()
   
   if self:attr("stunned") then
		self.stunned_counter = (self.stunned_counter or 0) + (self:attr("stun_immune") or 0) * 100
		if self.stunned_counter < 100 then
			self.energy.value = 0
		else
			-- We are saved for this turn
			self.stunned_counter = self.stunned_counter - 100
			game.logSeen(self, "%s temporarily fights the stun.", self.name:capitalize())
		end
	end

	if self.life <= 0 then self:die("fin") end
   
   -- Still enough energy to act ?
	if self.energy.value < game.energy_to_act then return false end

	return true
end

-- We use action points instead of energy to simulate multiple actions per turn
-- When action points hit 0 we end our turn
function _M:useActionPoints(value)
	local value = value or 20
	--BAD CODING Goes and gets the Uid of the party and refills the action points. Need to change every time a new entity is added, because the uid changes.
	local a = __uids[112]
	local b = __uids[12]
	--game.level.map:moveViewSurround(self.x, self.y, 8, 8)
	self:incActions(-value)
	if game.party.m_list[1].actions <= 0 and game.party.m_list[2].actions <= 0 then
		game.log("Out of action points")
		self:useEnergy()
		game.party:nextPlayer()
		--a.actions = 100
		--b.actions = 100
		self.changed = true
		return end
	-- Action points should never go below zero, but just in case
	if self:getActions() <= 0 then
		if self.actor == game.party.m_list[1] and game.party.m_list[2].actions <= 0 or self.actor == game.party.m_list[2] and game.party.m_list[1].actions <= 0 then
		game.party:nextPlayer()
		self.changed = true
		else
		self:useEnergy()
		end
	end
	self.changed = true
end

function _M:move(x, y, force)
	local moved = false
	local ox, oy = self.x, self.y
	
	if force or self:enoughEnergy() and self:getActions() >= 20 then
		moved = engine.Actor.move(self, x, y, force)
		if self.player then
		if not force and moved and (self.x ~= ox or self.y ~= oy) and not self.did_energy then self:useActionPoints(20) end
		end
		self:attr("moved_this_turn", 1)
	end
	--self.did_energy = nil
	--core.game.sleep(50)
	--self:setMoveAnim(self.x, self.y, 3, nil, util.getDir(x, y, self.x, self.y), 0.2)
	--game.level.map:moveViewSurround(self.x, self.y, 8, 8)
	return moved
end

function _M:tooltip()
	local desc = self.desc
   local name = self.name
   if game.player.inebriation > 60 then desc = self.drunk_desc end
   if game.player.inebriation > 60 and self ~= game.player then name = self.drunk_name end
   return ([[%s%s
#44dd44#Level: %d#LAST#
#CRIMSON#HP: %d (%d%%)#LAST#
#KHAKI#Stamina: %d (%d%%)#LAST#
Stats: Str:%d /  Aim:%d / Hea:%d / Spd:%d
%s
Stamina usage: #KHAKI#%d#LAST# per attack]]):format(
	self:getDisplayString(),
	name,
	self.level,
	self.life, self.life * 100 / self.max_life,
   self.stamina, self.stamina * 100 / self.max_stamina,
	self:getStr(),
	self:getAim(),
	self:getHea(),
   self:getSpd(),
	desc,
   self.stamina_cost
	)
end

function _M:onTakeHit(value, src)
	return value
end

function _M:die(src)
	engine.interface.ActorLife.die(self, src)

	-- Gives the killer some exp for the kill
	if src and src.gainExp then
		src:gainExp(self:worthExp(src))
	end
   

   -- Drop stuff
   for inven_id, inven in pairs(self.inven) do
      for i, o in ipairs(inven) do
         o.droppedBy = self.name
         game.level.map:addObject(self.x, self.y, o)
      end
   end
   self.inven = {}
   
	return true
end

function _M:levelup()
	self.max_life = self.max_life + 2

	self:incMaxStamina(5)

	-- Heal upon new level
	self.life = self.life + 2
	self.stamina = self.stamina + 5
      
end

--- Notifies a change of stat value
function _M:onStatChange(stat, v)
	if stat == self.STAT_HEALTH then
		self.max_life = self.max_life + 2*v
      self.life = self.life + 2*v
      self.max_stamina = self.max_stamina + v
      self.stamina = self.stamina + v
	end
   if stat == self.STAT_STRENGTH then
		self.max_stamina = self.max_stamina + 4*v
      self.stamina = self.stamina + 4*v
	end
   if stat == self.STAT_SPEED then
		self.energy.mod = self.energy.mod + 0.02*v
	end
   
end

function _M:attack(target)
	self:bumpInto(target)
end

function _M:added()
   engine.Actor.added(self)
   self:resetToFull()
end

-- Fill the Actor life and stamina
function _M:resetToFull()
   self.life = self.max_life
   self.stamina = self.max_stamina
   self.changed = true
end

--- Called before a talent is used
-- Check the actor can cast it
-- @param ab the talent (not the id, the table)
-- @return true to continue, false to stop
function _M:preUseTalent(ab, silent)
	if not self:enoughEnergy() then print("fail energy") return false end

	if ab.mode == "sustained" then
		if ab.sustain_stamina and self.max_stamina < ab.sustain_stamina and not self:isTalentActive(ab.id) then
			game.logPlayer(self, "You do not have enough stamina to activate %s.", ab.name)
			return false
		end
	else
		if ab.stamina and self:getStamina() < ab.stamina then
			game.logPlayer("I ain't got the energy for that.")
			return false
		end
	end

	return true
end

--- Called before a talent is used
-- Check if it must use a turn, mana, stamina, ...
-- @param ab the talent (not the id, the table)
-- @param ret the return of the talent action
-- @return true to continue, false to stop
function _M:postUseTalent(ab, ret)
	if not ret then return end

	--self:useActionPoints(20)

	if ab.mode == "sustained" then
		if not self:isTalentActive(ab.id) then
			if ab.sustain_stamina then
				self.max_stamina = self.max_stamina - ab.sustain_stamina
			end
		else
			if ab.sustain_stamina then
				self.max_stamina = self.max_stamina + ab.sustain_stamina
			end
		end
	else
		if ab.stamina then
			self:incStamina(-ab.stamina)
		end
	end

	return true
end

--- Return the full description of a talent
-- You may overload it to add more data (like power usage, ...)
function _M:getTalentFullDescription(t)
	local d = {}

	if t.mode == "passive" then d[#d+1] = "#6fff83#Use mode: #00FF00#Passive"
	elseif t.mode == "sustained" then d[#d+1] = "#6fff83#Use mode: #00FF00#Sustained"
	else d[#d+1] = "#6fff83#Use mode: #00FF00#Activated"
	end

	if t.stamina or t.sustain_stamina then d[#d+1] = "#6fff83#Stamina cost: #7fffd4#"..(t.stamina or t.sustain_stamina) end
	if self:getTalentRange(t) > 1 then d[#d+1] = "#6fff83#Range: #FFFFFF#"..self:getTalentRange(t)
	else d[#d+1] = "#6fff83#Range: #FFFFFF#melee/personal"
	end
	if t.cooldown then d[#d+1] = "#6fff83#Cooldown: #FFFFFF#"..t.cooldown end

	return table.concat(d, "\n").."\n#6fff83#Description: #FFFFFF#"..t.info(self, t)
end

--- How much experience is this actor worth
-- @param target to whom is the exp rewarded
-- @return the experience rewarded
function _M:worthExp(target)
	if not target.level or self.level < target.level - 3 then return 0 end

	local mult = 2
	if self.unique then mult = 6
	elseif self.egoed then mult = 3 end
	return self.level * mult * self.exp_worth
end

--- Can the actor see the target actor
-- This does not check LOS or such, only the actual ability to see it.<br/>
-- Check for telepathy, invisibility, stealth, ...
function _M:canSee(actor, def, def_pct)
	if not actor then return false, 0 end

	-- Check for stealth. Checks against the target cunning and level
	if actor:attr("stealth") and actor ~= self then
		local def = self.level / 2 + self:getSpd(25)
		local hit, chance = self:checkHit(def, actor:attr("stealth") + (actor:attr("inc_stealth") or 0), 0, 100)
		if not hit then
			return false, chance
		end
	end

	if def ~= nil then
		return def, def_pct
	else
		return true, 100
	end
end

--- Can the target be applied some effects
-- @param what a string describing what is being tried
function _M:canBe(what)
	if what == "poison" and rng.percent(100 * (self:attr("poison_immune") or 0)) then return false end
	if what == "cut" and rng.percent(100 * (self:attr("cut_immune") or 0)) then return false end
	if what == "confusion" and rng.percent(100 * (self:attr("confusion_immune") or 0)) then return false end
	if what == "blind" and rng.percent(100 * (self:attr("blind_immune") or 0)) then return false end
	if what == "stun" and rng.percent(100 * (self:attr("stun_immune") or 0)) then return false end
	if what == "fear" and rng.percent(100 * (self:attr("fear_immune") or 0)) then return false end
	if what == "knockback" and rng.percent(100 * (self:attr("knockback_immune") or 0)) then return false end
	if what == "instakill" and rng.percent(100 * (self:attr("instakill_immune") or 0)) then return false end
	return true
end

--- Remove some HP from an actor
-- If HP is reduced to 0 then remove from the level and call the die method.<br/>
-- When an actor dies its dead property is set to true, to wait until garbage collection deletes it
-- @return true/false if the actor died and the actual damage done
function _M:takeHit(value, src)
	if self.onTakeHit then value = self:onTakeHit(value, src) end
	self.life = self.life - value
	self.changed = true
	if self.life <= 0 then
      if self == game.player and src == game.player then game.log("The bullet passes through my skull, erupting in a crimson cloud that splatters against the grey walls.  The world seems to fade as I drop slowly to the ground.  A smile forms on my lips as it all goes dark.")
		elseif src == game.player then
         local desc = self.name
         if game.player.inebriation > 60 then desc = self.drunk_name end
         
         -- Enemy death messages
         if self.type == "animal" then
            if game.player.inebriation > 60 then game.logSeen(self, "The %s drops dead.", desc)
            else game.logSeen(self, "The %s collapses, vainly twitching before going still.", desc)
            end
         elseif self.subtype == "child" then
            game.logSeen(self, "The bloodied %s shudders and %s small body falls to the ground.  Ugh...", desc, self.his)
         elseif game.player.inebriation > 60 then game.logSeen(self, "The %s drops to the ground, bleedin' like a stuck pig.", desc)
         else game.logSeen(self, "The %s drops to the ground, bleeding profusely.  I watch the life fade from %s eyes.", desc, self.his)
         end
         
         if game.player.inebriation > 60 and self.type == "human" then game.player.inebriation = game.player.inebriation -1 end
         if game.player.inebriation > 60 and self.subtype == "child" then game.player.inebriation = game.player.inebriation -4 end
      elseif self == game.player then
         local desc = src.name
         if game.player.inebriation > 60 then desc = src.drunk_name end
         game.logSeen(self, "Reeling from the %s's blow I stumble and fall, and the world goes dark.", desc)
      end
		return self:die(src), value
	end
	return false, value
end
