-- Hellfire
-- Copyright (C) 2011 Sean "Chezzo" Osman sean@chezzo.com
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.


require "engine.class"
require "mod.class.Actor"
require "mod.class.Object"
require "engine.interface.GameTargeting"
require "engine.interface.PlayerRest"
require "engine.interface.PlayerRun"
require "engine.interface.PlayerMouse"
require "engine.interface.PlayerHotkeys"
require "engine.interface.PlayerSlide"
require "mod.class.interface.PartyDeath"
local Map = require "engine.Map"
local Dialog = require "engine.Dialog"
local ActorTalents = require "engine.interface.ActorTalents"
local LevelupStatsDialog = require "mod.dialogs.LevelupStatsDialog"
local DeathDialog = require "mod.dialogs.DeathDialog"
local Astar = require"engine.Astar"
local DirectPath = require"engine.DirectPath"
local DamageType = require "engine.DamageType"
local Particles = require "engine.Particles"

--- Defines the player
-- It is a normal actor, with some redefined methods to handle user interaction.<br/>
-- It is also able to run and rest and use hotkeys
module(..., package.seeall, class.inherit(
	mod.class.Actor,
	engine.interface.PlayerRest,
	engine.interface.PlayerRun,
	engine.interface.PlayerSlide,
	engine.interface.PlayerMouse,
	engine.interface.PlayerHotkeys,
	mod.class.interface.PartyDeath
))

-- Allow character registration even after birth
allow_late_uuid = true

function _M:init(t, no_default)
	t.display=t.display or '@'
	t.color_r=t.color_r or 75
	t.color_g=t.color_g or 54
	t.color_b=t.color_b or 33

	t.player = true
	t.type = t.type or "human"
	t.subtype = t.subtype or "player"
	t.faction = t.faction or "players"
	t.move_others=true
	t.lite = t.lite or 0
	--t.ai = "dumb_talented_simple"
	--t.ai_state = {talent_in=1, ai_move="move_astar"}

   t.inebriation = 50
   t.ineb_verbose = "Groggy"
   t.dreamcount = 0
   t.alcoholic = 0

   t.shadowed = false
   t.dream = true
   t.stamina_cost = 20
   t.unused_stats = 0
   t.flash = 1
   t.flashing = false
   t.desc=[[I'm Sam, a former construction worker in Hell. I'm in a world of hurt, and I gotta get outta here.]]
   
   t.his = "his"
   t.cries = "moans"
   t.he = "he"
      
	mod.class.Actor.init(self, t, no_default)
	engine.interface.PlayerHotkeys.init(self, t)

	self.descriptor = {}	
	self.died_times = self.died_times or {}
end

function _M:move(x, y, force)
	local ox, oy = self.x, self.y
	local moved = mod.class.Actor.move(self, x, y, force)

	if not force and ox == self.x and oy == self.y and self.doPlayerSlide then
		self.doPlayerSlide = nil
		tx, ty = self:tryPlayerSlide(x, y, false)
		if tx then moved = self:move(tx, ty, false) end
	end
	self.doPlayerSlide = nil
	--local x, y = self:getTarget(tg)
	--if not x then return end
	--local target = game.level.map(px, py, engine.Map.ACTOR)
    --if not target then return end
	  
	if moved then
		game.level.map:moveViewSurround(self.x, self.y, 8, 8)
      -- Highlight items on ground
		if self:getInven(self.INVEN_INVEN) then
			local i, nb = 1, 0
			local obj = game.level.map:getObject(self.x, self.y, i)
			while obj do
            nb = nb + 1
            i = i + 1
            obj = game.level.map:getObject(self.x, self.y, i)
			end
         
			if nb >= 2 then
				game.logSeen(self, "There's a few things on the ground here.")
			elseif nb == 1 then
				game.logSeen(self, "I see an item here: %s", game.level.map:getObject(self.x, self.y, 1):getName{do_color=true})
			end
		end
	end
	return moved
end

function _M:act()
	if not mod.class.Actor.act(self) then return end
   
   if game.party.shadowcount > 0 then
	   game.party.shadowcount = game.party.shadowcount - 1
	   
	   if game.party.shadowcount == 0 then
			-- The Shadow will get you!
			-- Find space
			local x, y = util.findFreeGrid(self.x, self.y, 1, true, {[Map.ACTOR]=true})
			if not x then
				game.party.shadowcount = 1
			else
				local NPC = require "mod.class.NPC"
				local m = NPC.new{
					type = "human",
					display = "@",
					image = "mushroom/npcs/unspeak.png",
					color=colors.DARK_GREY,
					name = "Tulu",
					desc = [[One of the things you let out of the box. Run for your life.]],
					ai = "dumb_talented_simple", ai_state = { talent_in=1, },
					stats = { str=10, aim=10, hea=10, spd=10 },
					combat_armor = 2,
					max_life = 250,
					combat = { dam=2, crit_chance = 0 },
					stamina_cost = 20,
					body = { INVEN = 40, MELEE=1, RANGED=1, BODY=1, HEAD=1 },
					copy = {
						max_level = 50,
						lite = 5,
						max_life = 100,
						image = "mushroom/pcs/Sam.png",
					  resolvers.equip{
							{type="melee", subtype="knife", name="blunt knife"},
							{type="gun", name="shotgun"},
						 },
					  resolvers.inventory{
							{type="ammo", subtype="shotgun", name="shells"},
							{type="ammo", subtype="shotgun", name="shells"},
							{type="ammo", subtype="shotgun", name="shells"},
							{type="ammo", subtype="shotgun", name="buckshot"},
							{type="gun", name="panzerfaust"},
						},	
					},
					talents = {
						[ActorTalents.T_KICK]=1,
						[ActorTalents.T_ACID_SPRAY]=1,
						[ActorTalents.T_BUCK_SHOT]=1,
						--[ActorTalents.T_MAKE_MAC]=1,
						--skull enhancement
						--[ActorTalents.T_BONE_MERGING]=1,
						--[ActorTalents.T_METALIC_ENHANCMENT]=1,
						--[ActorTalents.T_GEM_ENHANCMENT]=1,
						
					},
				}
				m.on_added_to_level = nil

				
				m:setTarget(game.player)
				m.energy.value = 0
				m.forceLevelup = function() end
				m.on_die = function()
					self.shadowed = nil
					self.win = true
					self.die(self)
				end
				game.zone:addEntity(game.level, m, "actor", x, y)
				--game.level.map:particleEmitter(x, y, 1, "teleport")
				self.shadowed = true
			end
	   end
	end
	
	if game.party.shadowcount == 50 then game.log("#ffff22#You feel relatively safe.") end
	if game.party.shadowcount == 45 then game.log("#ffff22#You feel a bit uneasy.") end
	if game.party.shadowcount == 40 then game.log("#ffff22#You notice a slight eldritch fog begin to surround you.") end
	if game.party.shadowcount == 35 then game.log("#ffff22#The fog is beginning to thicken.") end
	if game.party.shadowcount == 30 then game.log("#ffff22#You see a bolt of lightning crack off in the distance.") end
	if game.party.shadowcount == 25 then game.log("#ffff22#The fog is beginning to become dense.") end
	if game.party.shadowcount == 20 then game.log("#ffff22#You can barely see through the fog.") end
	if game.party.shadowcount == 15 then game.log("#ffff22#It is becoming hard to breathe.") end
	if game.party.shadowcount == 10 then game.log("#ffff22#You are filled with dread. Maybe you should get to the next level before something bad happens.") end
	if game.party.shadowcount == 8 then game.log("#ffff22#The fog is starting to burn into your flesh. Run.") end
	if game.party.shadowcount == 6 then game.log("#ffff22#The fog begins to crackle with lightning and arcane energy.") end
	if game.party.shadowcount == 4 then game.log("#ffff22#The lightning is growing and forming together much too close to you.") end
	if game.party.shadowcount == 3 then game.log("#ffff22#A portal cracks open a few feet from you just big enough for an eye to stare back at you from the other side. You feel your sanity bleeding out of you.") end
	if game.party.shadowcount == 2 then game.log("#ffff22#The portal widens and thousands of undescribable horrors reach at you. You barely avoid their grasp.") end
	if game.party.shadowcount == 1 then game.log("#ffff22#It widens further and one of the terrors is about to step through. Your life, death, and afterlife flash before your eyes.") end
	if game.party.shadowcount == 0 then game.log("#ffff22#A thing from the other side steps through the portal. If you value your continued existence, you will run.") end
	-- Funky shader things !
	self:updateMainShader()

	self.old_life = self.life
	
	-- Resting ? Running ? Otherwise pause
	if not self:restStep() and not self:runStep() and self.player then
		game.paused = true
	end
end

-- Precompute FOV form, for speed
local fovdist = {}
for i = 0, 30 * 30 do
	fovdist[i] = math.max((20 - math.sqrt(i)) / 14, 0.6)
end

function _M:playerFOV()
	-- Clean FOV before computing it
	game.level.map:cleanFOV()
	-- Compute both the normal and the lite FOV, using cache
	self:computeFOV(self.sight or 20, "block_sight", function(x, y, dx, dy, sqdist)
		game.level.map:apply(x, y, fovdist[sqdist])
	end, true, false, true)
	self:computeFOV(self.lite, "block_sight", function(x, y, dx, dy, sqdist) game.level.map:applyLite(x, y) end, true, true, true)
	
	-- For each entity, generate lite
	local uid, e = next(game.level.entities)
	while uid do
	if e ~= self and e.lite and e.lite > 0 and e.computeFOV then
	e:computeFOV(e.lite, "block_sight", function(x, y, dx, dy, sqdist) game.level.map:applyExtraLite(x, y, fovdist[sqdist]) end, true, true)
	end
	uid, e = next(game.level.entities, uid)
	end
	
end

--- Called before taking a hit, overload mod.class.Actor:onTakeHit() to stop resting and running
function _M:onTakeHit(value, src)
	self:runStop("I'm hit!")
	self:restStop("Agh, that hurt!")
   if self.inebriation > 40 then self.inebriation = self.inebriation - 0.5 end
	local ret = mod.class.Actor.onTakeHit(self, value, src)
	if self.life < self.max_life * 0.3 then
		local sx, sy = game.level.map:getTileToScreen(self.x, self.y)
		game.flyers:add(sx, sy, 30, (rng.range(0,2)-1) * 0.5, 2, "LOW HEALTH!", {255,0,0}, true)
	end
	return ret
end

function _M:die(src, deathdeath_note)
	self:runStop("died")
	self:restStop("died")
	
	return self:onPartyDeath(src, death_note)
end

function _M:setName(name)
	self.name = name
	game.save_name = name
end

--- Notify the player of available cooldowns
function _M:onTalentCooledDown(tid)
	local t = self:getTalentFromId(tid)

	local x, y = game.level.map:getTileToScreen(self.x, self.y)
	game.flyers:add(x, y, 30, -0.3, -3.5, ("%s available"):format(t.name:capitalize()), {0,255,00})
	game.log("#00ff00#Talent %s is ready to use.", t.name)
end

function _M:levelup()
	mod.class.Actor.levelup(self)

	local x, y = game.level.map:getTileToScreen(self.x, self.y)
	game.flyers:add(x, y, 30, 0.5, -2, "LEVEL UP!", {0,255,255})
	game.log("#55aaaa#Welcome to level %d.", self.level)
   
   self.unused_stats = self.unused_stats + 3
	local ds = LevelupStatsDialog.new(self)
	game:registerDialog(ds)
   
   self.max_life = 20 + 2*(self.level - 1) + 2*(self:getHea() - 10)
   self.max_stamina = 55 --[[+ self.level*5--]] + 3*self:getStr() + self:getHea()
   local speed = 1 + 0.02*(self:getSpd())
   self.energy.mod = speed
   self.energy.value = 0
end

--- Tries to get a target from the user
function _M:getTarget(typ)
	return game:targetGetForPlayer(typ)
end

--- Sets the current target
function _M:setTarget(target)
	return game:targetSetForPlayer(target)
end

local function spotHostiles(self)
	local seen = false
	-- Check for visible monsters, only see LOS actors, so telepathy wont prevent resting
	core.fov.calc_circle(self.x, self.y, game.level.map.w, game.level.map.h, 20, function(_, x, y) return game.level.map:opaque(x, y) end, function(_, x, y)
		local actor = game.level.map(x, y, game.level.map.ACTOR)
		if actor and self:reactionToward(actor) < 0 and self:canSee(actor) and game.level.map.seens(x, y) then seen = true end
	end, nil)
	return seen
end

--- Can we continue resting ?
-- We can rest if no hostiles are in sight, and if we need life/mana/stamina (and their regen rates allows them to fully regen)
function _M:restCheck()
	if spotHostiles(self) then return false, "There's something coming..." end

	-- Check resources, make sure they CAN go up, otherwise we will never stop
	if self:getStamina() < self:getMaxStamina() and self.stamina_regen > 0 then return true end
	if self.life < self.max_life and self.life_regen> 0 then return true end

	return false, "I'm ready to go."
end

-- Superload runStep to drop a game.paused = false in for Action Points to work
local previous_runStep = _M.runStep
function _M:runStep()
	game.paused = false
	return previous_runStep(self)
end


--- Can we continue running?
-- We can run if no hostiles are in sight, and if we no interesting terrains are next to us
function _M:runCheck()
	if self:getActions() >= 40 then return false, "Not enough action points." end
	if spotHostiles(self) then return false, "I see some one." end

	-- Notice any noticeable terrain
	local noticed = false
	self:runScan(function(x, y)
		-- Only notice interesting terrains
		local grid = game.level.map(x, y, Map.TERRAIN)
		if grid and grid.notice then noticed = "Hmm, where will this lead?" end
	end)
	if noticed then return false, noticed end

	self:playerFOV()
	
	return engine.interface.PlayerRun.runCheck(self)
end

--- Move with the mouse
-- We just feed our spotHostile to the interface mouseMove
function _M:mouseMove(tmx, tmy)
	local moves = self:getActions()/20

		local mx, my = core.mouse.get()
		local x, y = game.level.map:getMouseTile(mx, my)
		if not x or not y then return nil end
		Astar:init(game.level.map, game.player)
		if Astar:calc(self.x, self.y, x, y, false, nil, nil, false) == nil then return nil end
		--local moves_left = math.floor(core.fov.distance(self.x, self.y, x, y)) * 20
		if math.floor(core.fov.distance(self.x, self.y, x, y)) > moves then return nil end
		if math.floor(core.fov.distance(self.x, self.y, x, y)) == 2 then self:useActionPoints(20) end
		if math.floor(core.fov.distance(self.x, self.y, x, y)) == 3 then self:useActionPoints(40) end
		if math.floor(core.fov.distance(self.x, self.y, x, y)) == 4 then self:useActionPoints(60) end
		if math.floor(core.fov.distance(self.x, self.y, x, y)) == 5 then self:useActionPoints(80) end
		game.player:move(x, y, nil)
	--return engine.interface.PlayerMouse.mouseMove(self, tmx, tmy, spotHostiles)
end

--- Show combined equipment/inventory dialog
-- Overload to make it use the tooltip
function _M:showEquipInven(title, filter, action)
	local last = nil
	return mod.class.Actor.showEquipInven(self, title, filter, action, function(item)
		if item.last_display_x then
			game.tooltip_x, game.tooltip_y = {}, 1
			game.tooltip:displayAtMap(nil, nil, item.last_display_x, item.last_display_y, item.desc)
		end
	end)
end

function _M:doDrop(inven, item)
	self:dropFloor(inven, item, true, true)
	self:sortInven(inven)
	self:useActionPoints()
	self.changed = true
end

function _M:doWear(inven, item, o)
	self:removeObject(inven, item, true)
	local ro = self:wearObject(o, true, true)
	if ro then
		if type(ro) == "table" then self:addObject(inven, ro) end
	elseif not ro then
		self:addObject(inven, o)
	end
	self:sortInven()
    self:useActionPoints()
	self.changed = true
end

function _M:doTakeoff(inven, item, o)
	if self:takeoffObject(inven, item) then
		self:addObject(self.INVEN_INVEN, o)
	end
	self:sortInven()
    self:useActionPoints()
	self.changed = true
end

function _M:playerPickup()
	-- If 2 or more objects, display a pickup dialog, otherwise just picks up
	if game.level.map:getObject(self.x, self.y, 2) then
		
		local d d = self:showPickupFloor("Pick up", nil, function(o, item)
			self:pickupFloor(item, true)
			self.changed = true
			d.title = "Pick up"
			d:used()
		end)
	else
		self:pickupFloor(1, true)
		self:sortInven()
        self:useActionPoints()
	self.changed = true
	end
end

function _M:playerWear()
	local inven = self:getInven(self.INVEN_INVEN)
	local titleupdator = self:getEncumberTitleUpdator("Wield/wear object")
	self:showInventory(titleupdator(), inven, function(o)
		return o:wornInven() and self:getInven(o:wornInven())  and true or false
	end, function(o, item)
		self:doWear(inven, item, o)
	end)
end

function _M:playerTakeoff()
	local titleupdator = self:getEncumberTitleUpdator("Take off object")
	self:showEquipment(titleupdator(), nil, function(o, inven, item)
		self:doTakeoff(inven, item, o)
	end)
end

function _M:playerDrop()
	local inven = self:getInven(self.INVEN_INVEN)
	local titleupdator = self:getEncumberTitleUpdator("Drop object")
	self:showInventory(titleupdator(), inven, nil, function(o, item)
		self:doDrop(inven, item)
	end)
end

function _M:playerUseItem(object, item, inven)

	local use_fct = function(o, inven, item)
		if not o then return end
		local co = coroutine.create(function()
			self.changed = true

			local ret, id = o:use(self, nil, inven, item)
			if ret and ret == "destroy" then
				if o.multicharge and o.multicharge > 1 then
					o.multicharge = o.multicharge - 1
				else
					local _, del = self:removeObject(self:getInven(inven), item)
					if del then
						game.log("I'm out of %s.", o:getName{no_count=true, do_color=true})
					end
					self:sortInven(self:getInven(inven))
				end
				return true
			end

			self.changed = true
		end)
		local ok, ret = coroutine.resume(co)
		if not ok and ret then print(debug.traceback(co)) error(ret) end
		return true
	end

	if object and item then return use_fct(object, inven, item) end

	local titleupdator = self:getEncumberTitleUpdator("Use object")
	self:showEquipInven(titleupdator(),
		function(o)
			return o:canUseObject()
		end,
		use_fct,
		true
	)
end

function _M:playerShoot()
	local weapon = self:getInven("RANGED")[1]
	if not weapon or weapon.multicharge == 0 then
      game.log("I need a loaded weapon to shoot with.")
      return end

	local tg = {type="bolt", range=10}
	local x, y = self:getTarget(tg)
	if not x then return end
   weapon.multicharge = weapon.multicharge - 1
   game.log("I hold the gun steadily and shoot.")
	self:project(tg, x, y, DamageType.PHYSICAL, 25)
   self:project({type="ball",radius=5}, self.x, self.y, function(px, py)
      local target = game.level.map(px, py, engine.Map.ACTOR)
      if not target then return end
   end)
   	if self:getActions() <= 20 then
		  game.log("Not enough action points")
		  else
		  self:useActionPoints(40)
	end
end

--- Funky shader stuff
function _M:updateMainShader()
	if game.fbo_shader then
		-- Set shader HP warning
		if self.life ~= self.old_life then
			if self.life < self.max_life / 2 then game.fbo_shader:setUniform("hp_warning", 1 - (self.life / self.max_life))
			else game.fbo_shader:setUniform("hp_warning", 0) end
		end

		-- Colorize shader
		if self.shadowed then game.fbo_shader:setUniform("colorize", {0.1,0.1,0.1,0.75})
		else game.fbo_shader:setUniform("colorize", {0,0,0,0}) -- Disable
		end
		
		
		if game.party.shadowcount < 50 then game.fbo_shader:setUniform("colorize", {0.1,0.2,0.1,0.25}) end
		if game.party.shadowcount < 45 then game.fbo_shader:setUniform("colorize", {0.1,0.2,0.1,0.25}) end
		if game.party.shadowcount < 40 then game.fbo_shader:setUniform("colorize", {0.1,0.25,0.1,0.30}) end 
		if game.party.shadowcount < 35 then game.fbo_shader:setUniform("colorize", {0.1,0.25,0.1,0.30}) end
		if game.party.shadowcount < 30 then game.fbo_shader:setUniform("colorize", {0.1,0.3,0.1,0.35}) end
		if game.party.shadowcount < 27 then game.fbo_shader:setUniform("colorize", {0.1,0.3,0.1,0.35}) end
		if game.party.shadowcount < 25 then game.fbo_shader:setUniform("colorize", {0.1,0.35,0.1,0.45}) end
		if game.party.shadowcount < 22 then game.fbo_shader:setUniform("colorize", {0.1,0.35,0.1,0.45}) end
		if game.party.shadowcount < 20 then game.fbo_shader:setUniform("colorize", {0.1,0.4,0.1,0.55}) end
		if game.party.shadowcount < 17 then game.fbo_shader:setUniform("colorize", {0.1,0.45,0.1,0.55}) end
		if game.party.shadowcount < 15 then game.fbo_shader:setUniform("colorize", {0.1,0.5,0.1,0.65}) end
		if game.party.shadowcount < 12 then game.fbo_shader:setUniform("colorize", {0.1,0.55,0.1,0.65}) end
		if game.party.shadowcount < 10 then game.fbo_shader:setUniform("colorize", {0.1,0.6,0.1,0.80}) end
		if game.party.shadowcount < 5 then game.fbo_shader:setUniform("colorize", {0.1,0.65,0.1,0.80}) end
		if game.party.shadowcount < 2 then game.fbo_shader:setUniform("colorize", {0.1,0.70,0.1,0.9}) end
		if game.party.shadowcount < 0 then game.fbo_shader:setUniform("colorize", {0,0,0,0}) end

	end
end

function _M:getMac()
	local x, y = util.findFreeGrid(self.x, self.y, 1, true, {[Map.ACTOR]=true})
	if not x then
		self.log = ("Didn't work,")
	else
		local NPC = require "mod.class.NPC"
		local n = NPC.new{
			type = "human", subtype = "adult",
			display = "",
			image = "mushroom/pcs/mac.png",
			anim = {speed=0.15, loop=-1, max=16},
			color=colors.DARK_GREY,
			name = "Mac",
			desc = [[One of your friends. Probably your only friend now.]],
			--ai = "dumb_talented_simple", ai_state = { ai_move="move_dmap", talent_in=4,leader = self },
			faction = "players",
			stats = { str=10, aim=10, hea=10, spd=10 },
			combat_armor = 2,
			max_life = 100,
			
			lite = 5,
				combat = { dam=2, crit_chance = 0 },
			stamina_cost = 20,
			body = { INVEN = 40, MELEE=1, RANGED=1, BODY=1, HEAD=1 },
			talents = { [ActorTalents.T_KICK]=1,
			[ActorTalents.T_ACID_SPRAY]=1,
			[ActorTalents.T_BUCK_SHOT]=1,
			},
			inventory = { {type="ammo", subtype="shotgun", name="shells"},
            {type="ammo", subtype="shotgun", name="shells"},
            {type="ammo", subtype="shotgun", name="shells"},
			{type="ammo", subtype="shotgun", name="buckshot"},
			{type="gun", name="panzerfaust"},
			},
			hotkey = { 
			[1] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_CHANGE_LEVEL,
				},			
			[2] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_PICK_UP,
				},
			[3] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_END_TURN
				 
				},
			[4] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_KICK,
				},
			[5] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_NEXT_PLAYER
				 },
			[6] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_OPEN_INVENTORY
				},
			[7] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_RANGED_SHOT,
				},
			[8] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_BUCK_SHOT,
				},
			},
			save_hotkeys = true,
			hotkey_page = 1,
			move_others = true,
		}
		n.on_added_to_level = nil
		game.party:addMember(n, {
			control="full", type="human", title="Golem",
			orders = {target=true, leash=false, anchor=true, talents=true, behavior=true},
		})
		n:setTarget(game.player)
		n.energy.value = 1000
		n.actions = 100
		game.zone:addEntity(game.level, n, "actor", x, y)
	end
end

function _M:getSam()
	local x, y = util.findFreeGrid(self.x, self.y, 1, true, {[Map.ACTOR]=true})
	if not x then
		self.log = ("Didn't work,")
	else
		local NPC = require "mod.class.NPC"
		local n = NPC.new{
			type = "human", subtype = "adult",
			display = "",
			image = "mushroom/pcs/Sam.png",
			color=colors.DARK_GREY,
			name = "Sam",
			desc = [[I am Sam, a construction worker in Hell who only recently is completely screwed. I gotta get the hell outta here.]],
			--ai = "minion", ai_state = { ai_move="move_dmap", talent_in=4,leader = self },
			faction = self.faction,
			stats = { str=10, aim=10, hea=10, spd=10 },
			combat_armor = 0,
			lite = 5,
			max_life = 100,
				combat = { dam=2, crit_chance = 0 },
			stamina_cost = 20,
			body = { INVEN = 40, MELEE=1, RANGED=1, BODY=1, HEAD=1 },
		}
		n.on_added_to_level = nil
		game.party:addMember(n, {
			control="full", type="human", title="Sam",
			orders = {target=true, leash=true, anchor=true, talents=true, behavior=true},
		})
		n:setTarget(game.player)
		n.energy.value = 1000
		game.zone:addEntity(game.level, n, "actor", x, y)
	end
end

function _M:useThatEnergy()
self:useEnergy()
end

--BAD CODING! Goes and gets the global variable __uids to add 60 action points. Have to check the entity __uid before each release, as many entities are produced under these.
function _M:endOfTurn()
	game.log("Turn ending...")
	a = __uids[110]
	b = __uids[10]
	a.actions = 100
	b.actions = 100
	--self.player:useThatEnergy()
	self.changed = true
end
