-- Hellfire
-- Copyright (C) 2011 Sean "Chezzo" Osman sean@chezzo.com
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.

local Entity = require "engine.Entity"
local ActorTalents = require("engine.interface.ActorTalents")
local Textzone = require "engine.ui.Textzone"
local Map = require "engine.Map"
newTalentType{ type="role/combat", name = "combat", description = "Combat techniques" }

local oldNewTalent = Talents.newTalent


Talents.newTalent = function(self, t)
	assert(engine.interface.ActorTalents.talents_types_def[t.type[1]], "No talent category "..tostring(t.type[1]).." for talent "..t.name)
	if not t.image then
		t.image = "talents/"..(t.short_name or t.name):lower():gsub("[^a-z0-9_]", "_")..".png"
	end
	if fs.exists("/data/gfx/"..t.image) then t.display_entity = Entity.new{image=t.image, is_talent=true}
	else t.display_entity = Entity.new{image="talents/default.png", is_talent=true}
	end
	return oldNewTalent(self, t)
end


newTalent{
	name = "Kick",
	type = {"role/combat", 1},
	points = 1,
	cooldown = 6,
	power = 2,
	range = 1,
	action = function(self, t)
		local tg = {type="hit", range=self:getTalentRange(t)}
		local x, y, target = self:getTarget(tg)
		if not x or not y or not target then return nil end
		if math.floor(core.fov.distance(self.x, self.y, x, y)) > 1 then return nil end

		target:knockback(self.x, self.y, 5)
		return true
	end,
	info = function(self, t)
		return "Kick!"
	end,
}

newTalent{
	name = "Acid Spray",
	type = {"role/combat", 1},
	points = 1,
	cooldown = 6,
	power = 2,
	range = 6,
	action = function(self, t)
		local tg = {type="ball", range=self:getTalentRange(t), radius=1, talent=t}
		local x, y = self:getTarget(tg)
		if not x or not y then return nil end
		self:project(tg, x, y, DamageType.ACID, 5, {type="acid"})
		return true
	end,
	info = function(self, t)
		return "Zshhhhhhhhh!"
	end,
}

newTalent{
	name = "Ranged Shot",
	type = {"role/combat", 1},
	points = 1,
	cooldown = 1,
	power = 1,
	range = 5,
	action = function(self, t)
		local weapon = self:getInven("RANGED")[1]
	if not weapon or weapon.multicharge == 0 then
      game.log("I need a loaded weapon to shoot with.")
      return end
		  if self:getActions() <= 20 then
		  game.log("Not enough action points")
		  return
		  else
		  self:useActionPoints(40)
		  end
	local tg = {type="ball", range=self:getTalentRange(t), radius=0, talent=t}
	local x, y = self:getTarget(tg)
	if not x then return end
   weapon.multicharge = weapon.multicharge - 1
   game.log("I hold the gun steadily and shoot.")
  self:project(tg, x, y, DamageType.PHYSICAL, 10, {type="flame"})
      local target = game.level.map(px, py, engine.Map.ACTOR)
      if not target then return end
	end,
	info = function(self, t)
		return "Kerblam!"
	end,
}

newTalent{
	name = "Buck Shot",
	type = {"role/combat", 1},
	points = 1,
	cooldown = 1,
	power = 1,
	range = 5,
	action = function(self, t)
		local weapon = self:getInven("RANGED")[1]
	if not weapon or weapon.multicharge == 0 then
      game.log("I need a loaded weapon to shoot with.")
      return end
	if self:getActions() <= 20 then
		  game.log("Not enough action points")
		  return
		  else
		  self:useActionPoints(40)
		  end
	local tg = {type="cone", range=self:getTalentRange(t), radius=1, talent=t}
	local x, y = self:getTarget(tg)
	if not x then return end
  
   weapon.multicharge = weapon.multicharge - 1
   game.log("I hold the gun steadily and shoot.")
  self:project(tg, x, y, DamageType.PHYSICAL, 10, {type="dust_trail"})
      local target = game.level.map(px, py, engine.Map.ACTOR)
      if not target then return end
	end,
	info = function(self, t)
		return "BOOM!"
	end,
}

newTalent{
	name = "Open Inventory",
	type = {"role/combat", 1},
	desc = {"Or press I"},
	action = function(self, t)
		game.key:triggerVirtual("SHOW_INVENTORY")
		return true
	end,
	info = function(self, t)
		return "Opening inventory!"
	end,
}

newTalent{
	name = "Next Player",
	type = {"role/combat", 1},
	action = function(self, t)
		game.key:triggerVirtual("NEXT_PLAYER")
		return true
	end,
	info = function(self, t)
		return "Switching Players!"
	end,
}
	
newTalent{
	name = "End Turn",
	type = {"role/combat", 1},
	action = function(self, t)
		game.key:triggerVirtual("END_TURN")
		return true
	end,
	info = function(self, t)
		return "Ending Turn!"
	end,
}

newTalent{
	name = "Pick Up",
	type = {"role/combat", 1},
	action = function(self, t)
		game.key:triggerVirtual("PICKUP_FLOOR")
		return true
	end,
	info = function(self, t)
		return "Pick up!"
	end,
}

newTalent{
	name = "Make Mac",
	type = {"role/combat", 1},
	action = function(self, t)
		local NPC = require "mod.class.NPC"
		local n = NPC.new{
			type = "human", subtype = "adult",
			display = "",
			image = "mushroom/pcs/mac.png",
			anim = {speed=0.15, loop=-1, max=16},
			color=colors.DARK_GREY,
			name = "Mac",
			desc = [[One of your friends. Probably your only friend now.]],
			--ai = "dumb_talented_simple", ai_state = { ai_move="move_dmap", talent_in=4,leader = self },
			faction = "players",
			stats = { str=10, aim=10, hea=10, spd=10 },
			combat_armor = 2,
			max_life = 100,
			
			lite = 5,
				combat = { dam=2, crit_chance = 0 },
			stamina_cost = 20,
			body = { INVEN = 40, MELEE=1, RANGED=1, BODY=1, HEAD=1 },
			talents = { [ActorTalents.T_KICK]=1,
			[ActorTalents.T_ACID_SPRAY]=1,
			[ActorTalents.T_BUCK_SHOT]=1,
			},
			inventory = { {type="ammo", subtype="shotgun", name="shells"},
            {type="ammo", subtype="shotgun", name="shells"},
            {type="ammo", subtype="shotgun", name="shells"},
			{type="ammo", subtype="shotgun", name="buckshot"},
			{type="gun", name="panzerfaust"},
			},
			hotkey = { 
			[1] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_CHANGE_LEVEL,
				},			
			[2] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_PICK_UP,
				},
			[3] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_END_TURN
				 
				},
			[4] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_KICK,
				},
			[5] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_NEXT_PLAYER
				 },
			[6] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_OPEN_INVENTORY
				},
			[7] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_RANGED_SHOT,
				},
			[8] = {
				  [1] = "talent",
				  [2] = ActorTalents.T_BUCK_SHOT,
				},
			},
			save_hotkeys = true,
			hotkey_page = 1,
			move_others = true,
		}
		n.on_added_to_level = nil
		game.party:addMember(n, {
			control="full", type="human", title="Golem",
			orders = {target=true, leash=false, anchor=true, talents=true, behavior=true},
		})
		n:setTarget(game.player)
		n.energy.value = 1000
		n.actions = 100
		local x, y = util.findFreeGrid(self.x, self.y, 5, true, {[Map.ACTOR]=true})
		game.zone:addEntity(game.level, n, "actor", x, y)
		return true
	end,
	info = function(self, t)
		return "Made Mac!"
	end,
}

newTalent{
	name = "Change Level",
	type = {"role/combat", 1},
	action = function(self, t)
		game.key:triggerVirtual("CHANGE_LEVEL")
		return true
	end,
	info = function(self, t)
		return "Changing Level!"
	end,
}


--Uses talent movement/teleportation to override running and let player see where they have actions to go. 
--newTalent{
	--name = "Move",
	--type = {"role/combat", 1},
	--hide = true,
	--ignored_by_hotkeyautotalents = true,
	--action = function(self, t)
		--local moves = self:getActions()/20
		--local tg = {type="ball", range=moves, radius=0, talent=t}
		--local x, y = self:getTarget(tg)
		--if not x or not y then return nil end
		--local moves_left = math.floor(core.fov.distance(self.x, self.y, x, y)) * 20
		--if math.floor(core.fov.distance(self.x, self.y, x, y)) > moves then return nil end
		--if math.floor(core.fov.distance(self.x, self.y, x, y)) == 2 then self:useActionPoints(20) end
		--if math.floor(core.fov.distance(self.x, self.y, x, y)) == 3 then self:useActionPoints(40) end
		--if math.floor(core.fov.distance(self.x, self.y, x, y)) == 4 then self:useActionPoints(60) end
		--if math.floor(core.fov.distance(self.x, self.y, x, y)) == 5 then self:useActionPoints(80) end
		--game.player:move(x, y, nil)
		

		--return true
	--end,
	--info = function(self, t)
		--return "Changing Level!"
	--end,
--}

--newTalent{
--	name = "Make Mac",
--	type = {"role/combat", 1},
--	points = 1,
--	cooldown = 6,
--	range = 1,
--	action = function(self, t)
--	return require("mod.class.NPC").new{
--			type = "human", subtype = "adult",
--       	display = '', color=colors.RED, image = "mushroom/pcs/mac.png",
--			desc = [[Your best friend Mac.]],
--			faction = self.faction,
--			ai = "dumb_talented_simple", ai_state = { talent_in=2 },
--			name = "mac", exp_worth = 0,
--			max_level = 10,
--			lite = 5,
--			max_life = 25,

--			body = { INVEN = 80, MELEE = 1, RANGED = 1, BODY = 1, HEAD = 1 },
   
--		copy = {
--			max_level = 10,
--			lite = 5,
--			max_life = 25,
--			image = "mushroom/pcs/mac.png",
--		  resolvers.equip{
--				{type="melee", subtype="knife", name="blunt knife"},
--				{type="gun", name="revolver"},
--			 },
--		  resolvers.inventory{
--				{type="ammo", subtype="revolver", name="bullets"},
--				{type="ammo", subtype="revolver", name="bullets"},
--				{type="ammo", subtype="revolver", name="bullets"},
--				{type="stimulant", subtype="regeneration", name="stims"},
--				{type="liquor"},
--				{type="liquor"},
--			},

--		},
--		talents = {
--			[ActorTalents.T_KICK]=1,
--			[ActorTalents.T_ACID_SPRAY]=1,
--		},
--	}
--		local x, y = util.findFreeGrid(self.x, self.y, 5, true, {[Map.ACTOR]=true})
--				if not x then
--					game.logPlayer(self, "Not enough space to refit!")
--					return
--				end
--				game.zone:addEntity(game.level, undead, "actor", x, y)
--				return
--			end
--				info = function(self, t)
--			return "Hey Mac!"
--		end,
--	}
